@@ -1,10 +1,14 @@
+from __future__ import annotations
+
 import importlib
-import io
+import inspect as python_inspect
+import json
 import os
 import pprint
 import sys
 import warnings
 import webbrowser
+from contextlib import redirect_stdout
 from contextlib import suppress
 from pathlib import Path
 
@@ -13,18 +17,32 @@ from dynaconf import default_settings
 from dynaconf import LazySettings
 from dynaconf import loaders
 from dynaconf import settings as legacy_settings
+from dynaconf.base import Settings
 from dynaconf.loaders.py_loader import get_module
+from dynaconf.utils import prepare_json
 from dynaconf.utils import upperfy
 from dynaconf.utils.files import read_file
 from dynaconf.utils.functional import empty
+from dynaconf.utils.inspect import builtin_dumpers
+from dynaconf.utils.inspect import EnvNotFoundError
+from dynaconf.utils.inspect import inspect_settings
+from dynaconf.utils.inspect import KeyNotFoundError
+from dynaconf.utils.inspect import OutputFormatError
+from dynaconf.utils.inspect import print_debug_info
 from dynaconf.utils.parse_conf import parse_conf_data
+from dynaconf.utils.parse_conf import unparse_conf_data
 from dynaconf.validator import ValidationError
 from dynaconf.validator import Validator
 from dynaconf.vendor import click
 from dynaconf.vendor import toml
+from dynaconf.vendor import tomllib
 
+os.environ["PYTHONIOENCODING"] = "utf-8"
+
+CWD = None
+with suppress(FileNotFoundError):
+    CWD = Path.cwd()
 
-CWD = Path.cwd()
 EXTS = ["ini", "toml", "yaml", "json", "py", "env"]
 WRITERS = ["ini", "toml", "yaml", "json", "py", "redis", "vault", "env"]
 
@@ -33,11 +51,19 @@ ENC = default_settings.ENCODING_FOR_DYNACONF
 
 def set_settings(ctx, instance=None):
     """Pick correct settings instance and set it to a global variable."""
-
     global settings
 
     settings = None
 
+    _echo_enabled = ctx.invoked_subcommand not in [
+        "get",
+        "inspect",
+        "debug-info",
+        None,
+    ]
+    if "--json" in sys.argv:
+        _echo_enabled = False
+
     if instance is not None:
         if ctx.invoked_subcommand in ["init"]:
             raise click.UsageError(
@@ -48,25 +74,32 @@ def set_settings(ctx, instance=None):
     elif "FLASK_APP" in os.environ:  # pragma: no cover
         with suppress(ImportError, click.UsageError):
             from flask.cli import ScriptInfo  # noqa
+            from dynaconf import FlaskDynaconf
 
-            flask_app = ScriptInfo().load_app()
-            settings = flask_app.config
-            click.echo(
-                click.style(
-                    "Flask app detected", fg="white", bg="bright_black"
+            app_import_path = os.environ["FLASK_APP"]
+            flask_app = ScriptInfo(app_import_path).load_app()
+            settings = FlaskDynaconf(flask_app, **flask_app.config).settings
+            if _echo_enabled:
+                click.echo(
+                    click.style(
+                        "Flask app detected", fg="white", bg="bright_black"
+                    )
                 )
-            )
     elif "DJANGO_SETTINGS_MODULE" in os.environ:  # pragma: no cover
         sys.path.insert(0, os.path.abspath(os.getcwd()))
-        try:
-            # Django extension v2
-            from django.conf import settings  # noqa
+        import django  # noqa
 
-            settings.DYNACONF.configure()
-        except AttributeError:
-            settings = LazySettings()
-
-        if settings is not None:
+        django.setup()  # ensure django is setup to avoid AppRegistryNotReady
+        settings_module = import__django_settings(
+            os.environ["DJANGO_SETTINGS_MODULE"]
+        )
+        found_settings = python_inspect.getmembers(
+            settings_module,
+            lambda item: isinstance(item, (LazySettings, Settings)),
+        )
+        if found_settings:
+            settings = found_settings[0][1]
+        if settings is not None and _echo_enabled:
             click.echo(
                 click.style(
                     "Django app detected", fg="white", bg="bright_black"
@@ -74,8 +107,7 @@ def set_settings(ctx, instance=None):
             )
 
     if settings is None:
-
-        if instance is None and "--help" not in click.get_os_args():
+        if instance is None and "--help" not in sys.argv:
             if ctx.invoked_subcommand and ctx.invoked_subcommand not in [
                 "init",
             ]:
@@ -91,10 +123,22 @@ def set_settings(ctx, instance=None):
             settings = LazySettings()
 
 
+def import__django_settings(django_settings_module):
+    """Import the Django settings module from the string importable path."""
+    try:
+        with redirect_stdout(None):
+            module = importlib.import_module(django_settings_module)
+    except ImportError as e:
+        raise click.UsageError(e)
+    except FileNotFoundError:
+        return
+    return module
+
+
 def import_settings(dotted_path):
     """Import settings instance from python dotted path.
 
-    Last item in dotted path must be settings instace.
+    Last item in dotted path must be settings instance.
 
     Example: import_settings('path.to.settings')
     """
@@ -104,10 +148,14 @@ def import_settings(dotted_path):
         raise click.UsageError(
             f"invalid path to settings instance: {dotted_path}"
         )
+
     try:
-        module = importlib.import_module(module)
+        with redirect_stdout(None):
+            module = importlib.import_module(module)
     except ImportError as e:
         raise click.UsageError(e)
+    except FileNotFoundError:
+        return
     try:
         return getattr(module, name)
     except AttributeError as e:
@@ -158,7 +206,7 @@ def show_banner(ctx, param, value):
         return
     set_settings(ctx)
     click.echo(settings.dynaconf_banner)
-    click.echo("Learn more at: http://github.com/rochacbruno/dynaconf")
+    click.echo("Learn more at: http://github.com/dynaconf/dynaconf")
     ctx.exit()
 
 
@@ -242,7 +290,9 @@ def main(ctx, instance):
 @click.option("--django", default=os.environ.get("DJANGO_SETTINGS_MODULE"))
 @click.pass_context
 def init(ctx, fileformat, path, env, _vars, _secrets, wg, y, django):
-    """Inits a dynaconf project
+    """
+    Inits a dynaconf project.
+
     By default it creates a settings.toml and a .secrets.toml
     for [default|development|staging|testing|production|global] envs.
 
@@ -256,6 +306,19 @@ def init(ctx, fileformat, path, env, _vars, _secrets, wg, y, django):
     """
     click.echo("⚙️  Configuring your Dynaconf environment")
     click.echo("-" * 42)
+    if "FLASK_APP" in os.environ:  # pragma: no cover
+        click.echo(
+            "⚠️  Flask detected, you can't use `dynaconf init` "
+            "on a flask project, instead go to dynaconf.com/flask/ "
+            "for more information.\n"
+            "Or add the following to your app.py\n"
+            "\n"
+            "from dynaconf import FlaskDynaconf\n"
+            "app = Flask(__name__)\n"
+            "FlaskDynaconf(app)\n"
+        )
+        sys.exit(1)
+
     path = Path(path)
 
     if env is not None:
@@ -354,9 +417,7 @@ def init(ctx, fileformat, path, env, _vars, _secrets, wg, y, django):
 
     if settings_path:
         loader.write(settings_path, settings_data, merge=True)
-        click.echo(
-            f"🎛️  {settings_path.name} created to hold your settings.\n"
-        )
+        click.echo(f"🎛️  {settings_path.name} created to hold your settings.\n")
 
     if secrets_path:
         loader.write(secrets_path, secrets_data, merge=True)
@@ -364,15 +425,14 @@ def init(ctx, fileformat, path, env, _vars, _secrets, wg, y, django):
         ignore_line = ".secrets.*"
         comment = "\n# Ignore dynaconf secret files\n"
         if not gitignore_path.exists():
-            with io.open(str(gitignore_path), "w", encoding=ENC) as f:
+            with open(str(gitignore_path), "w", encoding=ENC) as f:
                 f.writelines([comment, ignore_line, "\n"])
         else:
             existing = (
-                ignore_line
-                in io.open(str(gitignore_path), encoding=ENC).read()
+                ignore_line in open(str(gitignore_path), encoding=ENC).read()
             )
             if not existing:  # pragma: no cover
-                with io.open(str(gitignore_path), "a+", encoding=ENC) as f:
+                with open(str(gitignore_path), "a+", encoding=ENC) as f:
                     f.writelines([comment, ignore_line, "\n"])
 
         click.echo(
@@ -401,6 +461,55 @@ def init(ctx, fileformat, path, env, _vars, _secrets, wg, y, django):
         )
 
 
+@main.command(name="get")
+@click.argument("key", required=True)
+@click.option(
+    "--default",
+    "-d",
+    default=empty,
+    help="Default value if settings doesn't exist",
+)
+@click.option(
+    "--env", "-e", default=None, help="Filters the env to get the values"
+)
+@click.option(
+    "--unparse",
+    "-u",
+    default=False,
+    help="Unparse data by adding markers such as @none, @int etc..",
+    is_flag=True,
+)
+def get(key, default, env, unparse):
+    """Returns the raw value for a settings key.
+
+    If result is a dict, list or tuple it is printed as a valid json string.
+    """
+    if env:
+        env = env.strip()
+    if key:
+        key = key.strip()
+
+    if env:
+        settings.setenv(env)
+
+    if default is not empty:
+        result = settings.get(key, default)
+    else:
+        try:
+            result = settings[key]
+        except KeyError:
+            click.echo("Key not found", nl=False, err=True)
+            sys.exit(1)
+
+    if unparse:
+        result = unparse_conf_data(result)
+
+    if isinstance(result, (dict, list, tuple)):
+        result = json.dumps(prepare_json(result), sort_keys=True, default=repr)
+
+    click.echo(result, nl=False)
+
+
 @main.command(name="list")
 @click.option(
     "--env", "-e", default=None, help="Filters the env to get the values"
@@ -441,9 +550,29 @@ def init(ctx, fileformat, path, env, _vars, _secrets, wg, y, django):
     default=False,
     help="Output file is flat (do not include [env] name)",
 )
-def _list(env, key, more, loader, _all=False, output=None, flat=False):
-    """Lists all user defined config values
-    and if `--all` is passed it also shows dynaconf internal variables.
+@click.option(
+    "--json",
+    "_json",
+    "-j",
+    is_flag=True,
+    default=False,
+    help="Prints out data serialized as JSON",
+)
+def _list(
+    env,
+    key,
+    more,
+    loader,
+    _all=False,
+    output=None,
+    flat=False,
+    _json=False,
+):
+    """
+    Lists user defined settings or all (including internal configs).
+
+    By default, shows only user defined. If `--all` is passed it also shows
+    dynaconf internal variables aswell.
     """
     if env:
         env = env.strip()
@@ -460,14 +589,15 @@ def _list(env, key, more, loader, _all=False, output=None, flat=False):
     if cur_env == "main":
         flat = True
 
-    click.echo(
-        click.style(
-            f"Working in {cur_env} environment ",
-            bold=True,
-            bg="bright_blue",
-            fg="bright_white",
+    if not _json:
+        click.echo(
+            click.style(
+                f"Working in {cur_env} environment ",
+                bold=True,
+                bg="bright_blue",
+                fg="bright_white",
+            )
         )
-    )
 
     if not loader:
         data = settings.as_dict(env=env, internal=_all)
@@ -493,14 +623,20 @@ def _list(env, key, more, loader, _all=False, output=None, flat=False):
         return f"{key}{data_type} {value}"
 
     if not key:
-        datalines = "\n".join(
-            format_setting(k, v)
-            for k, v in data.items()
-            if k not in data.get("RENAMED_VARS", [])
-        )
-        (click.echo_via_pager if more else click.echo)(datalines)
+        if not _json:
+            datalines = "\n".join(
+                format_setting(k, v)
+                for k, v in data.items()
+                if k not in data.get("RENAMED_VARS", [])
+            )
+            (click.echo_via_pager if more else click.echo)(datalines)
         if output:
-            loaders.write(output, data, env=not flat and cur_env)
+            loaders.write(output, prepare_json(data), env=not flat and cur_env)
+        if _json:
+            json_data = json.dumps(
+                prepare_json(data), sort_keys=True, default=repr
+            )
+            click.echo(json_data, nl=False)
     else:
         key = upperfy(key)
 
@@ -510,12 +646,19 @@ def _list(env, key, more, loader, _all=False, output=None, flat=False):
             value = empty
 
         if value is empty:
-            click.echo(click.style("Key not found", bg="red", fg="white"))
-            return
+            click.secho("Key not found", bg="red", fg="white", err=True)
+            sys.exit(1)
 
-        click.echo(format_setting(key, value))
+        if not _json:
+            click.echo(format_setting(key, value))
         if output:
-            loaders.write(output, {key: value}, env=not flat and cur_env)
+            loaders.write(
+                output, prepare_json({key: value}), env=not flat and cur_env
+            )
+        if _json:
+            click.echo(
+                json.dumps(prepare_json({key: value}), default=repr), nl=True
+            )
 
     if env:
         settings.setenv()
@@ -531,7 +674,7 @@ def _list(env, key, more, loader, _all=False, output=None, flat=False):
     default=None,
     help=(
         "key values to be written "
-        "e.g: `dynaconf write toml -e NAME=foo -e X=2"
+        "e.g: `dynaconf write toml -e NAME=foo -e X=2`"
     ),
 )
 @click.option(
@@ -542,7 +685,7 @@ def _list(env, key, more, loader, _all=False, output=None, flat=False):
     default=None,
     help=(
         "secret key values to be written in .secrets "
-        "e.g: `dynaconf write toml -s TOKEN=kdslmflds -s X=2"
+        "e.g: `dynaconf write toml -s TOKEN=kdslmflds -s X=2`"
     ),
 )
 @click.option(
@@ -564,13 +707,12 @@ def _list(env, key, more, loader, _all=False, output=None, flat=False):
 )
 @click.option("-y", default=False, is_flag=True)
 def write(to, _vars, _secrets, path, env, y):
-    """Writes data to specific source"""
+    """Writes data to specific source."""
     _vars = split_vars(_vars)
     _secrets = split_vars(_secrets)
     loader = importlib.import_module(f"dynaconf.loaders.{to}_loader")
 
     if to in EXTS:
-
         # Lets write to a file
         path = Path(path)
 
@@ -637,12 +779,14 @@ def write(to, _vars, _secrets, path, env, y):
     "--path", "-p", default=CWD, help="defaults to current directory"
 )
 def validate(path):  # pragma: no cover
-    """Validates Dynaconf settings based on rules defined in
-    dynaconf_validators.toml"""
+    """
+    Validates Dynaconf settings based on provided rules.
+
+    Rules should be defined in dynaconf_validators.toml
+    """
     # reads the 'dynaconf_validators.toml' from path
     # for each section register the validator for specific env
     # call validate
-
     path = Path(path)
 
     if not str(path).endswith(".toml"):
@@ -652,7 +796,31 @@ def validate(path):  # pragma: no cover
         click.echo(click.style(f"{path} not found", fg="white", bg="red"))
         sys.exit(1)
 
-    validation_data = toml.load(open(str(path)))
+    # parse validator file
+    try:  # try tomlib first
+        validation_data = tomllib.load(open(str(path), "rb"))
+    except UnicodeDecodeError:  # fallback to legacy toml (TBR in 4.0.0)
+        warnings.warn(
+            "TOML files should have only UTF-8 encoded characters. "
+            "starting on 4.0.0 dynaconf will stop allowing invalid chars.",
+        )
+        validation_data = toml.load(
+            open(str(path), encoding=default_settings.ENCODING_FOR_DYNACONF),
+        )
+    except tomllib.TOMLDecodeError as e:
+        click.echo(
+            click.style(
+                f"Error parsing TOML: {e}. Maybe it should be quoted.",
+                fg="white",
+                bg="red",
+            )
+        )
+        sys.exit(1)
+
+    # guarantee there is an environment
+    validation_data = {k.lower(): v for k, v in validation_data.items()}
+    if not validation_data.get("default"):
+        validation_data = {"default": validation_data}
 
     success = True
     for env, name_data in validation_data.items():
@@ -660,7 +828,8 @@ def validate(path):  # pragma: no cover
             if not isinstance(data, dict):  # pragma: no cover
                 click.echo(
                     click.style(
-                        f"Invalid rule for parameter '{name}'",
+                        f"Invalid rule for parameter '{name}'"
+                        "(this will be skipped)",
                         fg="white",
                         bg="yellow",
                     )
@@ -689,5 +858,98 @@ def validate(path):  # pragma: no cover
         sys.exit(1)
 
 
+INSPECT_FORMATS = list(builtin_dumpers.keys())
+
+
+@main.command()
+@click.option("--key", "-k", help="Filters result by key.")
+@click.option(
+    "--env",
+    "-e",
+    help="Filters result by environment on --report-mode=inspect.",
+    default=None,
+)
+@click.option(
+    "--format",
+    "-f",
+    help="The output format.",
+    default="json",
+    type=click.Choice(INSPECT_FORMATS),
+)
+@click.option(
+    "--old-first",
+    "new_first",
+    "-s",
+    help="Invert history sorting to 'old-first' on --report-mode=inspect.",
+    default=True,
+    is_flag=True,
+)
+@click.option(
+    "--limit",
+    "history_limit",
+    "-n",
+    default=None,
+    type=int,
+    help="Limits how many history entries are shown on --report-mode=inspect.",
+)
+@click.option(
+    "--all",
+    "_all",
+    "-a",
+    default=False,
+    is_flag=True,
+    help="Show dynaconf internal settings?",
+)
+@click.option(
+    "--report-mode",
+    "report_mode",
+    "-m",
+    default="inspect",
+    type=click.Choice(["inspect", "debug"]),
+)
+@click.option(
+    "-v",
+    "--verbose",
+    count=True,
+    help="Increase verbosity of the output on --report-mode=debug.",
+)
+def inspect(
+    key, env, format, new_first, history_limit, _all, report_mode, verbose
+):  # pragma: no cover
+    """
+    Inspect the loading history of the given settings instance.
+
+    Filters by key and environment, otherwise shows all.
+    """
+
+    if report_mode == "debug":
+        print_debug_info(
+            settings,
+            dumper=format,
+            verbosity=verbose,
+            key=key,
+        )
+        click.echo()
+    elif report_mode == "inspect":
+        try:
+            inspect_settings(
+                settings,
+                key=key,
+                env=env or None,
+                dumper=format,
+                new_first=new_first,
+                include_internal=_all,
+                history_limit=history_limit,
+                print_report=True,
+            )
+            click.echo()
+        except (KeyNotFoundError, EnvNotFoundError, OutputFormatError) as err:
+            click.echo(err)
+            sys.exit(1)
+    else:
+        click.echo("Invalid report mode")
+        sys.exit(1)
+
+
 if __name__ == "__main__":  # pragma: no cover
     main()
